unit GsvZoomer;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls;

type
  TGsvZoomerPressedType = (zptNone, zptDown, zptUp,
                           zptPageDown, zptPageUp,
                           zptSlider, zptSliderDown, zptSliderUp);
  TGsvZoomerNotify = procedure(Sender: TObject; Position, Range: Integer)
                     of object;

  TGsvZoomer = class(TCustomControl)
  private
    FGlyphs:             TBitmap;
    FTimer:              TTimer;

    FModelPosition:      Integer;
    FModelRange:         Integer;
    FModelMinRange:      Integer;
    FModelMaxRange:      Integer;
    FShowSlider:         Boolean;

    FOnChanged:          TGsvZoomerNotify;

    FTimeStartRepeat:    Cardinal;
    FRepeatDelay:        Cardinal;
    FSliderX:            Integer;
    FSliderWidth:        Integer;
    FPressedType:        TGsvZoomerPressedType;
    FPressedX:           Integer;
    FPressedSliderX:     Integer;
    FPressedSliderWidth: Integer;
    FPressedPosition:    Integer;
    FPressedRange:       Integer;

    procedure SetModelPosition(Value: Integer);
    procedure SetModelRange(Value: Integer);
    procedure SetModelMinRange(Value: Integer);
    procedure SetModelMaxRange(Value: Integer);

    function  ModelToSlider: Boolean;
    procedure VerifyData;
    procedure VerifySliderWidth;

    function  GetPressedType(X: Integer): TGsvZoomerPressedType;
    function  GetLineSize: Integer;
    function  GetPageSize: Integer;

    procedure OnTimer(Sender: TObject);
    procedure DoChanged;

    procedure ChangePositionByPressedType;

  protected
    procedure Paint; override;
    procedure Resize; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState;
              X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState;
              X, Y: Integer); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;

    procedure   SetData(aMinRange, aMaxRange, aPosition, aRange: Integer);
    procedure   HideSlider;

    property    Position: Integer read FModelPosition write SetModelPosition;
    property    Range: Integer read FModelRange write SetModelRange;
    property    MinRange: Integer read FModelMinRange write SetModelMinRange;
    property    MaxRange: Integer read FModelMaxRange write SetModelMaxRange;

  published
    property Align;
    property Enabled;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Visible;
    property OnChanged: TGsvZoomerNotify read FOnChanged write FOnChanged;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Gsv', [TGsvZoomer]);
end;

{$R GsvZoomerBitmaps.res}

const
  WS  = 6; //      
  HS  = 1; //         
  XSR = 2; //     
  YSR = 2; //     
  GCR = 1; //    
  WR  = 2; //  
  FSliderMinWidth = 2 * WS + 3 * WR + 4 * GCR + 2;

{ TGsvZoomer }

procedure TGsvZoomer.SetModelPosition(Value: Integer);
begin
  if (Value + FModelRange) > FModelMaxRange then
    Value := FModelMaxRange - FModelRange;
  if Value < 0 then
    Value := 0;
  if FModelPosition <> Value then begin
    FModelPosition := Value;
    if ModelToSlider then
      Invalidate;
  end;
end;

procedure TGsvZoomer.SetModelRange(Value: Integer);
var
  p: Integer;
begin
  p := FModelPosition;
  if (p + Value) > FModelMaxRange then
    Value := FModelMaxRange - p;
  if Value > FModelMaxRange then
    Value := FModelMaxRange;
  if Value < FModelMinRange then
    Value := FModelMinRange;
  if (p + Value) > FModelMaxRange then
    p := FModelMaxRange - Value;
  if p < 0 then
    p := 0;
  if (FModelRange <> Value) or (FModelPosition <> p) then begin
    FModelRange    := Value;
    FModelPosition := p;
    if ModelToSlider then
      Invalidate;
  end;
end;

procedure TGsvZoomer.SetModelMinRange(Value: Integer);
var
  p, r: Integer;
begin
  p := FModelPosition;
  r := FModelRange;
  if Value < 1 then
    Value := 1;
  if Value > (FModelMaxRange - 1) then
    Value := FModelMaxRange - 1;
  if r < Value then
    r := Value;
  if (p + r) > FModelMaxRange then
    p := FModelMaxRange - r;
  if p < 0 then
    p := 0;
  if (FModelMinRange <> Value) or
     (FModelPosition <> p) or
     (FModelRange <> r) then
  begin
    FModelMinRange := Value;
    FModelPosition := p;
    FModelRange    := r;
    if ModelToSlider then
      Invalidate;
  end;
end;

procedure TGsvZoomer.SetModelMaxRange(Value: Integer);
var
  p, r: Integer;
begin
  p := FModelPosition;
  r := FModelRange;
  if Value < 2 then
    Value := 2;
  if Value < (FModelMinRange + 1) then
    Value := FModelMinRange + 1;
  if r > Value then
    r := Value;
  if (p + r) > Value then
    p := Value - r;
  if p < 0 then
    p := 0;
  if (FModelMaxRange <> Value) or
     (FModelPosition <> p) or
     (FModelRange <> r) then
  begin
    FModelMaxRange := Value;
    FModelPosition := p;
    FModelRange    := r;
    if ModelToSlider then
      Invalidate;
  end;
end;

function TGsvZoomer.ModelToSlider: Boolean;
var
  x, w, mw, mp, mr: Integer;
begin
  mw := Width - 2 * Height;
  if mw > 0 then
    mr := MulDiv(FModelMaxRange, FSliderMinWidth, mw)
  else
    mr := FModelMinRange;
  if FModelRange > mr then
    mr := FModelRange;
  mp := FModelMaxRange - mr;
  x  := FSliderX;
  w  := FSliderWidth;
  FSliderWidth := MulDiv(mr, mw, FModelMaxRange);
  VerifySliderWidth;
  FSliderX := Height + MulDiv(FModelPosition, mw - FSliderWidth, mp);
  if (FSliderX + FSliderWidth) > (Width - Height) then
    FSliderX := Width - Height - FSliderWidth;
  if FSliderX < Height then
    FSliderX := Height;
  Result := (x <> FSliderX) or (w <> FSliderWidth);
end;

procedure TGsvZoomer.VerifyData;
begin
  if FModelMinRange < 1 then
    FModelMinRange := 1;
  if FModelMaxRange < (FModelMinRange + 1) then
    FModelMaxRange := FModelMinRange + 1;
  if FModelRange > FModelMaxRange then
    FModelRange := FModelMaxRange;
  if FModelRange < FModelMinRange then
    FModelRange := FModelMinRange;
  if (FModelPosition + FModelRange) > FModelMaxRange then
    FModelPosition := FModelMaxRange - FModelRange;
  if FModelPosition < 0 then
    FModelPosition := 0;
end;

procedure TGsvZoomer.VerifySliderWidth;
var
  mw, mr: Integer;
begin
  mw := Width - 2 * Height;
  if FSliderWidth > mw then
    FSliderWidth := mw;
  if FSliderWidth < FSliderMinWidth then
    FSliderWidth := FSliderMinWidth;
  mr := MulDiv(FModelMinRange, mw, FModelMaxRange);
  if FSliderWidth < mr then
    FSliderWidth := mr;
end;

function TGsvZoomer.GetPressedType(X: Integer): TGsvZoomerPressedType;
const
  G = 4;
begin
  if X < Height then
    Result := zptDown
  else if X >= (Width - Height) then
    Result := zptUp
  else if X < FSliderX then
    Result := zptPageDown
  else if X >= (FSliderX + FSliderWidth) then
    Result := zptPageUp
  else if X < (FSliderX + G) then
    Result := zptSliderDown
  else if X >= (FSliderX + FSliderWidth - G) then
    Result := zptSliderUp
  else
    Result := zptSlider;
end;

function TGsvZoomer.GetLineSize: Integer;
begin
  Result := FModelRange div 10;
  if Result = 0 then
    Result := 1;
end;

function TGsvZoomer.GetPageSize: Integer;
begin
  Result := FModelRange;
end;

procedure TGsvZoomer.OnTimer(Sender: TObject);
begin
  if GetTickCount > FTimeStartRepeat then
    ChangePositionByPressedType;
end;

procedure TGsvZoomer.DoChanged;
begin
  if Assigned(FOnChanged) and FShowSlider then
    FOnChanged(Self, FModelPosition, FModelRange);
end;

procedure TGsvZoomer.ChangePositionByPressedType;
var
  p: Integer;
begin
  p := FModelPosition;
  case FPressedType of
    zptDown:     SetModelPosition(FModelPosition - GetLineSize);
    zptUp:       SetModelPosition(FModelPosition + GetLineSize);
    zptPageDown: SetModelPosition(FModelPosition - GetPageSize);
    zptPageUp:   SetModelPosition(FModelPosition + GetPageSize);
  end;
  if p <> FModelPosition then
    DoChanged;
end;

procedure TGsvZoomer.Paint;
var
  r:                TRect;
  w, h:             Integer;
  xbL, xbR, xL, xR: Integer;
  x, xC, YCR:       Integer;
  flags:            Cardinal;
  {c:                TColor;
  y:                Integer;}

  procedure DrawTopLine(aX0, aX1: Integer);
  begin
    Canvas.Pen.Color := clBtnFace;
    Canvas.MoveTo(0, 0);
    Canvas.LineTo(w, 0);
  end;

begin
  w   := Width;
  h   := Height;
  xbL := h;
  xbR := w - h;
  xL  := FSliderX;
  xR  := xL + FSliderWidth;
  xC  := xL + ((xR - xL) div 2) - 1;
  YCR := (h - (3 * WR + 2 * GCR - 1)) div 2;

  if not FShowSlider then begin
    r := Rect(0, 0, w, h);
    Canvas.Brush.Bitmap := FGlyphs;
    Canvas.FillRect(r);
    Canvas.Brush.Bitmap := nil;
    DrawTopLine(0, w);
    Exit;
  end;

  //   
  if w >= h then begin
    flags := DFCS_SCROLLLEFT;
    if FPressedType = zptDown then
      flags := flags or DFCS_PUSHED or DFCS_FLAT;
    r := Rect(0, 0, xbL, h);
    DrawFrameControl(Canvas.Handle, r, DFC_SCROLL, flags);
  end;

  //   
  if w >= (2 * h) then begin
    flags := DFCS_SCROLLRIGHT;
    if FPressedType = zptUp then
      flags := flags or DFCS_PUSHED or DFCS_FLAT;
    r := Rect(xbR, 0, w, h);
    DrawFrameControl(Canvas.Handle, r, DFC_SCROLL, flags);
  end;

  //   
  if xL > xbL then begin
    r := Rect(xbL, 0, xL, h);
    if FPressedType = zptPageDown then
      Canvas.Brush.Color := clBtnShadow
    else
      Canvas.Brush.Bitmap := FGlyphs;
    Canvas.FillRect(r);
    DrawTopLine(xbL, xL);
  end;

  //   
  if xR < xbR then begin
    r := Rect(xbR, 0, xR, h);
    if FPressedType = zptPageUp then
      Canvas.Brush.Color := clBtnShadow
    else if Canvas.Brush.Bitmap = nil then
      Canvas.Brush.Bitmap := FGlyphs;
    Canvas.FillRect(r);
    DrawTopLine(xbR, xR);
  end;
  Canvas.Brush.Bitmap := nil;

  if (w - 2 * h) > FSliderMinWidth then begin
    //   
    r := Rect(xL, 0, xR, h);
    Canvas.Brush.Color := clRed;
    Canvas.FillRect(r);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT or BF_MIDDLE);
    //   
    r := Rect(xL + WS, HS, xR - WS, h - HS);
    DrawEdge(Canvas.Handle, r, BDR_SUNKENOUTER, BF_RECT);
    //  
    {c := $C0C0C0;
    for y := r.Bottom - 2 downto r.Top + 1 do begin
      Canvas.Pen.Color := c;
      Canvas.MoveTo(xL + WS, y);
      Canvas.LineTo(xR - WS - 1, y);
      c := c + $060606;
      if c > $FFFFFF then
        c := $FFFFFF;
    end;}

    //  
    x := xL + XSR;
    r := Rect(x, YSR, x + WR, h - YSR);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
    x := xR - XSR - WR - 1;
    r := Rect(x, YSR, x + WR, h - YSR);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);

    //  
    x := xC;
    r := Rect(x, YCR, x + WR, h - YCR);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
    x := xC - GCR - WR;
    r := Rect(x, YCR, x + WR, h - YCR);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
    x := xC + WR + GCR;
    r := Rect(x, YCR, x + WR, h - YCR);
    DrawEdge(Canvas.Handle, r, BDR_RAISEDINNER, BF_RECT);
  end;
end;

procedure TGsvZoomer.Resize;
begin
  ModelToSlider;
end;

procedure TGsvZoomer.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  FPressedType        := GetPressedType(X);
  FPressedX           := X;
  FPressedSliderX     := FSliderX;
  FPressedSliderWidth := FSliderWidth;
  FPressedPosition    := FModelPosition;
  FPressedRange       := FModelRange;
  FTimer.Enabled      := (FPressedType >= zptDown) and (FPressedType <= zptPageUp);
  FTimeStartRepeat    := GetTickCount + FRepeatDelay;
  Invalidate;
end;

procedure TGsvZoomer.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  px, pw, mw, d: Integer;

  procedure ChangeX;
  var
    mp, mx: Integer;
  begin
    mp := FModelMaxRange - FPressedRange;
    mx := mw - FSliderWidth;
    if mx > 0 then
      FPressedPosition := MulDiv(FSliderX - Height, mp, mx)
    else
      FPressedPosition := 0;
  end;

  procedure ChangeW;
  begin
    if mw > 0 then
      FPressedRange := MulDiv(FSliderWidth, FModelMaxRange, mw);
  end;

  procedure ChangeD;
  var
    p, r: Integer;
  begin
    p              := FModelPosition;
    r              := FModelRange;
    FModelPosition := FPressedPosition;
    FModelRange    := FPressedRange;
    VerifyData;
    if (p <> FModelPosition) or (r <> FModelRange) then
      DoChanged;
  end;

begin
  mw := Width - 2 * Height;
  if FPressedType = zptNone then begin
    if GetPressedType(X) > zptSlider then
      Cursor := crSizeWE
    else
      Cursor := crDefault;
  end
  else if mw >= FSliderMinWidth then begin
    px := FSliderX;
    pw := FSliderWidth;
    if FPressedType = zptSlider then begin
      FSliderX := FPressedSliderX - (FPressedX - X);
      if (FSliderX + FSliderWidth) > (Width - Height) then
        FSliderX := Width - Height - FSliderWidth;
      if FSliderX < Height then
        FSliderX := Height;
      ChangeX;
      ChangeD;
    end
    else if FPressedType = zptSliderDown then begin
      FSliderX := FPressedSliderX - (FPressedX - X);
      if FSliderX < Height then
        FSliderX := Height;
      d := FPressedSliderX + FPressedSliderWidth - FSliderMinWidth;
      if FSliderX > d then
        FSliderX := d;
      FSliderWidth := FPressedSliderX + FPressedSliderWidth - FSliderX;
      VerifySliderWidth;
      ChangeX;
      ChangeW;
      ChangeD;
    end
    else if FPressedType = zptSliderUp then begin
      FSliderWidth := FPressedSliderWidth + (X - FPressedX);
      d := Width - Height - FSliderX;
      if FSliderWidth > d then
        FSliderWidth := d;
      VerifySliderWidth;  
      ChangeW;
      ChangeD;
    end;
    if (px <> FSliderX) or (pw <> FSliderWidth) then
      Invalidate;
  end;
end;

procedure TGsvZoomer.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  FTimer.Enabled := False;
  if FPressedType = GetPressedType(X) then
    ChangePositionByPressedType;
  FPressedType := zptNone;
  Cursor := crDefault;
  Invalidate;
end;

constructor TGsvZoomer.Create(AOwner: TComponent);
var
  d: Cardinal;
begin
  inherited Create(AOwner);
  ControlStyle   := [csCaptureMouse, csClickEvents, csOpaque, csFixedHeight];
  TabStop        := False;
  DoubleBuffered := True;
  Width          := 200;
  Height         := GetSystemMetrics(SM_CYHSCROLL);

  FModelPosition := 0;
  FModelMaxRange := 100;
  FModelMinRange := 10;
  FModelRange    := 50;
  ModelToSlider;

  FGlyphs := TBitmap.Create;
  FGlyphs.LoadFromResourceName(HInstance, 'GSVZOOMERGLYPHS');

  FTimer         := TTimer.Create(Self);
  FTimer.Enabled := False;
  FTimer.OnTimer := OnTimer;
  SystemParametersInfo(SPI_GETKEYBOARDDELAY, 0, @d, 0);
  if d > 3 then
    d := 3;
  FRepeatDelay := 250 * (d + 1);
  SystemParametersInfo(SPI_GETKEYBOARDSPEED, 0, @d, 0);
  if d > 31 then
    d := 31;
  FTimer.Interval := 400 - 12 * d;
end;

destructor TGsvZoomer.Destroy;
begin
  FGlyphs.Free;
  inherited Destroy;
end;

procedure TGsvZoomer.SetData(aMinRange, aMaxRange, aPosition, aRange: Integer);
begin
  FModelMinRange := aMinRange;
  FModelMaxRange := aMaxRange;
  FModelPosition := aPosition;
  FModelRange    := aRange;
  FShowSlider    := True;
  VerifyData;
  ModelToSlider;
  Invalidate;
end;

procedure TGsvZoomer.HideSlider;
begin
  FShowSlider := False;
  Invalidate;
end;

end.
